# vim:sw=4:ts=4:et:
"""Python Ring Doorbell module."""

from __future__ import annotations

import logging
import time
from itertools import chain
from typing import TYPE_CHECKING, Any, ClassVar

from ring_doorbell import RingEvent
from ring_doorbell.chime import RingChime
from ring_doorbell.doorbot import RingDoorBell
from ring_doorbell.exceptions import RingError
from ring_doorbell.group import RingLightGroup
from ring_doorbell.other import RingOther
from ring_doorbell.stickup_cam import RingStickUpCam

from .const import (
    API_URI,
    API_VERSION,
    DEVICES_ENDPOINT,
    DINGS_ENDPOINT,
    GROUPS_ENDPOINT,
    INTERCOM_KINDS,
    NEW_SESSION_ENDPOINT,
)

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping, Sequence

    from ring_doorbell.auth import Auth
    from ring_doorbell.generic import RingGeneric

_logger = logging.getLogger(__name__)


class Ring:
    """A Python Abstraction object to Ring Door Bell."""

    def __init__(self, auth: Auth) -> None:
        """Initialize the Ring object."""
        self.auth: Auth = auth
        self.session = None
        self.subscription = None
        self.devices_data: dict[str, dict[int, dict[str, Any]]] = {}
        self._devices: RingDevices | None = None
        self.chime_health_data = None
        self.doorbell_health_data = None
        self.dings_data: dict[Any, Any] = {}
        self.push_dings_data: list[RingEvent] = []
        self.groups_data: dict[str, dict[str, Any]] = {}
        self.init_loop = None
        self.session_refresh_time: float | None = None

    async def async_update_data(self) -> None:
        """Update all data."""
        await self._async_update_data()

    async def _async_update_data(self) -> None:
        if self.session is None:
            await self.async_create_session()

        await self.async_update_devices()

        await self.async_update_dings()

        await self.async_update_groups()

    def _add_event_to_dings_data(self, ring_event: RingEvent) -> None:
        # Purge expired push_dings
        now = time.time()
        self.push_dings_data = [
            re for re in self.push_dings_data if now < re.now + re.expires_in
        ]
        self.push_dings_data.append(ring_event)

    async def async_create_session(self) -> None:
        """Create a new Ring session."""
        session_post_data = {
            "device": {
                "hardware_id": self.auth.get_hardware_id(),
                "metadata": {
                    "api_version": API_VERSION,
                    "device_model": self.auth.get_device_model(),
                },
                "os": "android",
            }
        }
        resp = await self._async_query(
            NEW_SESSION_ENDPOINT,
            method="POST",
            json=session_post_data,
        )
        self.session = resp.json()
        self.session_refresh_time = time.monotonic()

    async def async_update_devices(self) -> None:
        """Update device data."""
        if self.session is None:
            await self.async_create_session()

        resp = await self._async_query(DEVICES_ENDPOINT)
        data: dict[Any, Any] = resp.json()

        # Index data by device ID.
        self.devices_data = {
            device_type: {obj["id"]: obj for obj in devices}
            for device_type, devices in data.items()
        }

    async def async_update_dings(self) -> None:
        """Update dings data."""
        if self.session is None:
            await self.async_create_session()

        resp = await self._async_query(DINGS_ENDPOINT)
        self.dings_data = resp.json()

    async def async_update_groups(self) -> None:
        """Update groups data."""
        if self.session is None:
            await self.async_create_session()
        # Get all locations
        locations = set()
        devices = self.devices()
        for device_type in devices:
            for dev in devices[device_type]:
                if dev.location_id is not None:
                    locations.add(dev.location_id)

        # Query for groups
        self.groups_data = {}
        for location in locations:
            resp = await self._async_query(GROUPS_ENDPOINT.format(location))
            data = resp.json()
            if data["device_groups"]:
                for group in data["device_groups"]:
                    self.groups_data[group["device_group_id"]] = group

    async def async_query(  # noqa: PLR0913
        self,
        url: str,
        method: str = "GET",
        extra_params: dict[str, Any] | None = None,
        data: bytes | None = None,
        json: dict[Any, Any] | None = None,
        timeout: float | None = None,
        base_uri: str = API_URI,
    ) -> Auth.Response:
        """Query data from Ring API."""
        if self.session is None:
            await self.async_create_session()
        return await self._async_query(
            url, method, extra_params, data, json, timeout, base_uri
        )

    async def _async_query(  # noqa: PLR0913
        self,
        url: str,
        method: str = "GET",
        extra_params: dict[str, Any] | None = None,
        data: bytes | None = None,
        json: dict[Any, Any] | None = None,
        timeout: float | None = None,
        base_uri: str = API_URI,
    ) -> Auth.Response:
        _logger.debug(
            "url: %s\nmethod: %s\njson: %s\ndata: %s\n extra_params: %s",
            url,
            method,
            json,
            data,
            extra_params,
        )
        return await self.auth.async_query(
            base_uri + url,
            method=method,
            extra_params=extra_params,
            data=data,
            json=json,
            timeout=timeout,
        )

    def devices(self) -> RingDevices:
        """Get all devices."""
        if not self._devices:
            self._devices = RingDevices(self, self.devices_data)
        return self._devices

    def get_device_list(self) -> Sequence[RingGeneric]:
        """Get a combined list of all devices."""
        devices = self.devices()
        return list(
            chain(
                devices["doorbots"],
                devices["authorized_doorbots"],
                devices["stickup_cams"],
                devices["chimes"],
                devices["other"],
            )
        )

    def get_device_by_name(self, device_name: str) -> RingGeneric | None:
        """Return a device using it's name."""
        all_devices = self.get_device_list()
        names_to_idx = {device.name: idx for (idx, device) in enumerate(all_devices)}
        return (
            None
            if device_name not in names_to_idx
            else all_devices[names_to_idx[device_name]]
        )

    def get_video_device_by_name(self, device_name: str) -> RingDoorBell | None:
        """Return a device using it's name."""
        video_devices = self.video_devices()
        names_to_idx = {device.name: idx for (idx, device) in enumerate(video_devices)}
        return (
            None
            if device_name not in names_to_idx
            else video_devices[names_to_idx[device_name]]
        )

    def get_device_by_api_id(self, device_api_id: int) -> RingGeneric | None:
        """Return a device using it's id."""
        all_devices = self.get_device_list()
        api_id_to_idx = {
            device.device_api_id: idx for (idx, device) in enumerate(all_devices)
        }
        return (
            None
            if device_api_id not in api_id_to_idx
            else all_devices[api_id_to_idx[device_api_id]]
        )

    def video_devices(self) -> Sequence[RingDoorBell]:
        """Get all devices."""
        devices = self.devices()
        return list(
            chain(devices.doorbots, devices.authorized_doorbots, devices.stickup_cams)
        )

    def groups(self) -> Mapping[str, RingLightGroup]:
        """Get all groups."""
        groups = {}

        for group_id in self.groups_data:
            groups[group_id] = RingLightGroup(self, group_id)

        return groups

    def active_alerts(self) -> Sequence[RingEvent]:
        """Get active alerts."""
        now = time.time()
        # Purge expired push_dings
        self.push_dings_data = [
            re for re in self.push_dings_data if now < re.now + re.expires_in
        ]
        # Get unique id dictionary
        alerts: dict[tuple[int, int, str], RingEvent] = {}
        for re in self.push_dings_data:
            key = (re.doorbot_id, re.id, re.kind)
            if key not in alerts or re.now > alerts[key].now:
                alerts[key] = re

        for ding_data in self.dings_data:
            expires_at = ding_data.get("now") + ding_data.get("expires_in")

            if now < expires_at:
                re = RingEvent(
                    id=ding_data["id"],
                    doorbot_id=ding_data["doorbot_id"],
                    device_name=ding_data["doorbot_description"],
                    device_kind=ding_data["device_kind"],
                    now=ding_data["now"],
                    expires_in=ding_data["expires_in"],
                    kind=ding_data["kind"],
                    state=ding_data["state"],
                )
                key = (re.doorbot_id, re.id, re.kind)
                if key not in alerts or re.now > alerts[key].now:
                    alerts[key] = re

        return list(alerts.values())

    DEPRECATED_API_QUERIES: ClassVar = {
        "update_devices",
        "update_data",
        "update_dings",
        "update_groups",
        "create_session",
        "query",
    }

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            """Get a deprecated attribute or raise an error."""
            if name in self.DEPRECATED_API_QUERIES:
                return self.auth._dep_handler.get_api_query(self, name)  # noqa: SLF001
            msg = f"{self.__class__.__name__} has no attribute {name!r}"
            raise AttributeError(msg)


class RingDevices:
    """Class to represent collection of devices."""

    def __init__(
        self, ring: Ring, devices_data: dict[str, dict[int, dict[str, Any]]]
    ) -> None:
        """Initialise the devices from the api response."""
        self._stickup_cams: list[RingStickUpCam] = []
        self._chimes: list[RingChime] = []
        self._doorbots: list[RingDoorBell] = []
        self._authorized_doorbots: list[RingDoorBell] = []
        self._other: list[RingOther] = []

        for device_type, devices in devices_data.items():
            if device_type == "stickup_cams":
                self._stickup_cams = [
                    RingStickUpCam(ring, device_id) for device_id in devices
                ]
            if device_type == "chimes":
                self._chimes = [RingChime(ring, device_id) for device_id in devices]
            if device_type == "doorbots":
                self._doorbots = [
                    RingDoorBell(ring, device_id) for device_id in devices
                ]
            if device_type == "authorized_doorbots":
                self._authorized_doorbots = [
                    RingDoorBell(ring, device_id, shared=True) for device_id in devices
                ]
            if device_type == "other":
                self._other = [
                    RingOther(ring, device_id, shared=True)
                    for device_id, device in devices.items()
                    if (device_kind := device.get("kind"))
                    and device_kind in INTERCOM_KINDS
                ]
        self._all_devices = {
            device.id: device
            for device in chain(
                self._stickup_cams,
                self._chimes,
                self._doorbots,
                self._authorized_doorbots,
                self._other,
            )
        }

    def __getitem__(self, device_type: str) -> Sequence[RingGeneric]:
        """Get a generic device by type."""
        if device_type == "stickup_cams":
            return self._stickup_cams
        if device_type == "chimes":
            return self._chimes
        if device_type == "doorbots":
            return self._doorbots
        if device_type == "authorized_doorbots":
            return self._authorized_doorbots
        if device_type == "other":
            return self._other
        if device_type == "intercoms":
            return self._other
        msg = f"Invalid device_type {device_type}"
        raise RingError(msg)

    def __iter__(self) -> Iterator[str]:
        """Device type iterator."""
        return iter(
            ["stickup_cams", "chimes", "doorbots", "authorized_doorbots", "other"]
        )

    @property
    def stickup_cams(self) -> Sequence[RingStickUpCam]:
        """The stickup cams."""
        return self._stickup_cams

    @property
    def chimes(self) -> Sequence[RingChime]:
        """The chimes."""
        return self._chimes

    @property
    def doorbots(self) -> Sequence[RingDoorBell]:
        """The doorbots."""
        return self._doorbots

    @property
    def authorized_doorbots(self) -> Sequence[RingDoorBell]:
        """The authorized_doorbots."""
        return self._authorized_doorbots

    @property
    def doorbells(self) -> Sequence[RingDoorBell]:
        """The doorbells, i.e. doorbots and authorized_doorbots combined."""
        return self._doorbots + self._authorized_doorbots

    @property
    def other(self) -> Sequence[RingOther]:
        """The other devices, i.e. intercoms."""
        return self._other

    @property
    def all_devices(self) -> Sequence[RingGeneric]:
        """All devices combined."""
        return list(self._all_devices.values())

    @property
    def video_devices(self) -> Sequence[RingDoorBell]:
        """The video devices, i.e. doorbells and stickup_cams."""
        return [*self._doorbots, *self._authorized_doorbots, *self._stickup_cams]

    def get_device(self, device_api_id: int) -> RingGeneric:
        """Get device by api id."""
        if device := self._all_devices.get(device_api_id):
            return device
        msg = f"device with id {device_api_id} not found"
        raise RingError(msg)

    def get_doorbell(self, device_api_id: int) -> RingDoorBell:
        """Get doorbell by api id."""
        if (
            (device := self._all_devices.get(device_api_id))
            and isinstance(device, RingDoorBell)
            and not issubclass(device.__class__, RingDoorBell)
        ):
            return device
        msg = f"doorbell with id {device_api_id} not found"
        raise RingError(msg)

    def get_stickup_cam(self, device_api_id: int) -> RingStickUpCam:
        """Get stickup_cam by api id."""
        if (device := self._all_devices.get(device_api_id)) and isinstance(
            device, RingStickUpCam
        ):
            return device
        msg = f"stickup_cam with id {device_api_id} not found"
        raise RingError(msg)

    def get_chime(self, device_api_id: int) -> RingChime:
        """Get chime by api id."""
        if (device := self._all_devices.get(device_api_id)) and isinstance(
            device, RingChime
        ):
            return device
        msg = f"chime with id {device_api_id} not found"
        raise RingError(msg)

    def get_other(self, device_api_id: int) -> RingOther:
        """Get other device by api id."""
        if (device := self._all_devices.get(device_api_id)) and isinstance(
            device, RingOther
        ):
            return device
        msg = f"other device with id {device_api_id} not found"
        raise RingError(msg)

    def get_video_device(self, device_api_id: int) -> RingDoorBell:
        """Get video capable device by api id."""
        if (device := self._all_devices.get(device_api_id)) and isinstance(
            device, RingDoorBell
        ):
            return device
        msg = f"video capable device with id {device_api_id} not found"
        raise RingError(msg)

    def __str__(self) -> str:
        """Get string representation of devices."""
        d = {dev_type: self.__getitem__(dev_type) for dev_type in self.__iter__()}
        return "{" + "\n".join(f"{k!r}: {v!r}," for k, v in d.items()) + "}"

    def __repr__(self) -> str:
        """Return repr of devices."""
        d = {dev_type: self.__getitem__(dev_type) for dev_type in self.__iter__()}
        return repr(d)



# vim:sw=4:ts=4:et:
"""Python Ring Auth Class."""

from __future__ import annotations

import uuid
from asyncio import TimeoutError
from functools import cached_property
from json import loads as json_loads
from typing import TYPE_CHECKING, Any, Callable, ClassVar

from aiohttp import BasicAuth, ClientError, ClientResponseError, ClientSession
from oauthlib.common import urldecode
from oauthlib.oauth2 import (
    LegacyApplicationClient,
    MissingTokenError,
    OAuth2Error,
    TokenExpiredError,
)

from ring_doorbell.const import NAMESPACE_UUID, TIMEOUT, OAuth
from ring_doorbell.exceptions import (
    AuthenticationError,
    Requires2FAError,
    RingError,
    RingTimeout,
)
from ring_doorbell.util import _DeprecatedSyncApiHandler


class Auth:
    """A Python Auth class for Ring."""

    def __init__(
        self,
        user_agent: str,
        token: dict[str, Any] | None = None,
        token_updater: Callable[[dict[str, Any]], None] | None = None,
        hardware_id: str | None = None,
        *,
        http_client_session: ClientSession | None = None,
    ) -> None:
        """Initialise the auth class.

        :type token: Optional[Dict[str, str]]
        :type token_updater: Optional[Callable[[str], None]]
        """
        self.user_agent = user_agent

        if hardware_id:
            self.hardware_id = hardware_id
        else:
            # Generate a UUID that will stay the same
            # for this physical device to prevent
            # multiple auth entries in ring.com
            self.hardware_id = str(
                uuid.uuid5(uuid.UUID(NAMESPACE_UUID), str(uuid.getnode()) + user_agent)
            )

        self.device_model = "ring-doorbell:" + user_agent
        self.token_updater = token_updater
        self._token: dict[str, Any] = token or {}
        self._local_session: ClientSession | None = None
        self.http_client_session = http_client_session
        self._oauth_client = LegacyApplicationClient(
            client_id=OAuth.CLIENT_ID, token=token
        )
        self._auth = BasicAuth(OAuth.CLIENT_ID, "")

    @property
    def _session(self) -> ClientSession:
        if self.http_client_session:
            return self.http_client_session
        if self._local_session is None:
            self._local_session = ClientSession()
        return self._local_session

    async def async_fetch_token(
        self, username: str, password: str, otp_code: str | None = None
    ) -> dict[str, Any]:
        """Fetch initial token with username/password & 2FA.

        :type username: str
        :type password: str
        :type otp_code: str.
        """
        headers = {"User-Agent": self.user_agent, "hardware_id": self.hardware_id}

        if otp_code:
            headers["2fa-support"] = "true"
            headers["2fa-code"] = otp_code

        try:
            body = self._oauth_client.prepare_request_body(
                username, password, scope=OAuth.SCOPE, include_client_id=True
            )
            data = dict(urldecode(body))
            resp = await self._session.request(
                "POST",
                OAuth.ENDPOINT,
                data=data,
                headers=headers,
                auth=self._auth,
            )
            async with resp:
                text = await resp.text()
            self._token = self._oauth_client.parse_request_body_response(
                text, scope=OAuth.SCOPE
            )
        except MissingTokenError as ex:
            raise Requires2FAError from ex
        except OAuth2Error as ex:
            raise AuthenticationError(ex) from ex

        if self.token_updater is not None:
            self.token_updater(self._token)

        return self._token

    async def async_refresh_tokens(self) -> dict[str, Any]:
        """Refresh the auth tokens."""
        try:
            headers = {
                "Accept": "application/json",
                "Content-Type": ("application/x-www-form-urlencoded;charset=UTF-8"),
            }
            body = self._oauth_client.prepare_refresh_body(
                refresh_token=self._token["refresh_token"]
            )
            data = dict(urldecode(body))
            resp = await self._session.request(
                "POST", OAuth.ENDPOINT, data=data, headers=headers, auth=self._auth
            )
            async with resp:
                text = await resp.text()
            self._token = self._oauth_client.parse_request_body_response(
                text, scope=OAuth.SCOPE
            )
        except OAuth2Error as ex:
            raise AuthenticationError(ex) from ex

        if self.token_updater is not None:
            self.token_updater(self._token)

        return self._token

    def get_hardware_id(self) -> str:
        """Get hardware ID."""
        return self.hardware_id

    def get_device_model(self) -> str:
        """Get device model."""
        return self.device_model

    async def async_close(self) -> None:
        """Close aiohttp session."""
        session = self._local_session
        self._local_session = None
        if session:
            await session.close()

    class Response:
        """Class for returning responses."""

        def __init__(self, content: bytes, status_code: int) -> None:
            """Initialise thhe repsonse class."""
            self.content = content
            self.status_code = status_code

        @property
        def text(self) -> str:
            """Response as text."""
            return self.content.decode()

        def json(self) -> Any:
            """Response as loaded json."""
            return json_loads(self.text)

    async def async_query(  # noqa: C901, PLR0913
        self,
        url: str,
        method: str = "GET",
        extra_params: dict[str, Any] | None = None,
        data: bytes | None = None,
        json: dict[Any, Any] | None = None,
        timeout: float | None = None,
        *,
        raise_for_status: bool = True,
    ) -> Auth.Response:
        """Query data from Ring API."""
        if timeout is None:
            timeout = TIMEOUT

        params = {}
        if extra_params:
            params.update(extra_params)

        kwargs: dict[str, Any] = {
            "params": params,
            "timeout": timeout,
        }
        headers = {"User-Agent": self.user_agent, "hardware_id": self.get_hardware_id()}
        # Ring servers started requiring a null json value for PUT requests in 2024-10
        if json is not None or method == "PUT":
            kwargs["json"] = json
            headers["Content-Type"] = "application/json"

        try:
            try:
                url, headers, data = self._oauth_client.add_token(
                    url,
                    http_method=method,
                    body=data,
                    headers=headers,
                )

                resp = await self._session.request(
                    method, url, headers=headers, data=data, **kwargs
                )
            except TokenExpiredError:
                self._token = await self.async_refresh_tokens()
                url, headers, data = self._oauth_client.add_token(
                    url,
                    http_method=method,
                    body=data,
                    headers=headers,
                )
                resp = await self._session.request(
                    method, url, headers=headers, data=data, **kwargs
                )
        except AuthenticationError:
            raise  # refresh_tokens will return this error if not valid
        except TimeoutError as ex:
            msg = f"Timeout error during query of url {url}: {ex}"
            raise RingTimeout(msg) from ex
        except ClientError as ex:
            msg = f"aiohttp Client error during query of url {url}: {ex}"
            raise RingError(msg) from ex
        except Exception as ex:
            msg = f"Unknown error during query of url {url}: {ex}"
            raise RingError(msg) from ex

        async with resp:
            if resp.status == 401:
                # Check whether there's an issue with the token grant
                self._token = await self.async_refresh_tokens()

            if raise_for_status:
                try:
                    resp.raise_for_status()
                except ClientResponseError as ex:
                    msg = (
                        f"HTTP error with status code {resp.status} "
                        f"during query of url {url}: {ex}"
                    )
                    raise RingError(msg) from ex

            response_data = await resp.read()
        return Auth.Response(response_data, resp.status)

    @cached_property
    def _dep_handler(self) -> _DeprecatedSyncApiHandler:
        return _DeprecatedSyncApiHandler(self)

    DEPRECATED_API_QUERIES: ClassVar = {
        "fetch_token",
        "refresh_tokens",
        "close",
        "query",
    }

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            """Get a deprecated attribute or raise an error."""
            if name in self.DEPRECATED_API_QUERIES:
                return self._dep_handler.get_api_query(self, name)
            msg = f"{self.__class__.__name__} has no attribute {name!r}"
            raise AttributeError(msg)



"""Module for common utility functions."""

from __future__ import annotations

import asyncio
import datetime
import logging
from contextlib import suppress
from functools import update_wrapper
from threading import Lock
from typing import TYPE_CHECKING, Any, Callable
from warnings import warn

from typing_extensions import ParamSpec, TypeVar

from ring_doorbell.exceptions import RingError

if TYPE_CHECKING:
    from collections.abc import Coroutine

    from .auth import Auth
    from .generic import RingGeneric
    from .group import RingLightGroup
    from .listen.eventlistener import RingEventListener
    from .ring import Ring

    _T = TypeVar(
        "_T", bound=Auth | Ring | RingGeneric | RingLightGroup | RingEventListener
    )
    _R = TypeVar("_R")
    _P = ParamSpec("_P")

_logger = logging.getLogger(__name__)


def parse_datetime(datetime_str: str) -> datetime.datetime:
    """Parse a datetime string into a datetime object.

    Ring api has inconsistent datetime string patterns.
    """
    # Check if the datetime string contains a period which precedes 'Z',
    # indicating microseconds
    if "." in datetime_str and datetime_str.endswith("Z"):
        # String contains microseconds and ends with 'Z'
        format_str = "%Y-%m-%dT%H:%M:%S.%fZ"
    else:
        # String does not contain microseconds, should end with 'Z'
        # Could be updated to handle other formats
        format_str = "%Y-%m-%dT%H:%M:%SZ"
    try:
        res = datetime.datetime.strptime(datetime_str, format_str).replace(
            tzinfo=datetime.timezone.utc
        )
    except ValueError:
        _logger.exception(
            "Unable to parse datetime string %s, defaulting to now time", datetime_str
        )
        res = datetime.datetime.now(datetime.timezone.utc)
    return res


class _DeprecatedSyncApiHandler:
    def __init__(self, auth: Auth) -> None:
        self.auth = auth
        self._sync_lock = Lock()

    async def run_and_close_session(
        self,
        async_method: Callable[_P, Coroutine[Any, Any, _R]],
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> _R:
        try:
            self._sync_lock.acquire()
            res = await async_method(*args, **kwargs)
        finally:
            with suppress(Exception):
                await self.auth.async_close()
            self._sync_lock.release()

        return res

    @staticmethod
    def check_no_loop(classname: str, method_name: str) -> None:
        current_loop = None
        with suppress(RuntimeError):
            current_loop = asyncio.get_running_loop()
        if current_loop:
            msg = (
                f"You cannot call deprecated sync function {classname}.{method_name} "
                "from within a running event loop."
            )
            raise RingError(msg)

    def get_api_query(
        self,
        class_instance: _T,
        method_name: str,
    ) -> Any:
        """Return deprecated sync api query attribute."""
        classname = type(class_instance).__name__

        def _deprecated_sync_function(
            async_func: Callable[_P, Coroutine[Any, Any, _R]],
        ) -> Callable[_P, _R]:
            def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
                self.check_no_loop(classname, method_name)
                msg = (
                    f"{classname}.{method_name} is deprecated, use "
                    f"{classname}.{async_method_name}"
                )
                warn(msg, DeprecationWarning, stacklevel=1)
                return asyncio.run(
                    self.run_and_close_session(async_func, *args, **kwargs)
                )

            return update_wrapper(wrapper, async_func)

        async_method_name = f"async_{method_name}"
        async_method = getattr(class_instance, async_method_name)
        return _deprecated_sync_function(async_method)

    def get_api_property(
        self,
        class_instance: _T,
        method_name: str,
    ) -> Any:
        """Return deprecated sync api property value."""
        classname = type(class_instance).__name__
        self.check_no_loop(classname, method_name)
        async_method_name = f"async_get_{method_name}"
        msg = (
            f"{classname}.{method_name} is deprecated, use "
            f"{classname}.{async_method_name}"
        )
        warn(msg, DeprecationWarning, stacklevel=1)
        async_method = getattr(class_instance, async_method_name)
        return asyncio.run(self.run_and_close_session(async_method))

    def set_api_property(
        self,
        class_instance: _T,
        property_name: str,
        value: Any,
    ) -> None:
        """Set sync api property value."""
        classname = type(class_instance).__name__

        self.check_no_loop(classname, property_name)
        async_method_name = f"async_set_{property_name}"
        msg = (
            f"{classname}.{property_name} is deprecated, use "
            f"{classname}.{async_method_name}"
        )
        warn(msg, DeprecationWarning, stacklevel=1)
        async_method = getattr(class_instance, async_method_name)
        asyncio.run(self.run_and_close_session(async_method, value))



"""Module for ring events."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, NamedTuple


@dataclass
class RingEvent:
    """Class for ring events."""

    id: int
    doorbot_id: int
    device_name: str
    device_kind: str
    now: float
    expires_in: float
    kind: str
    state: str
    is_update: bool = False

    def __getitem__(self, key: str) -> Any:
        """Get a value by string."""
        return getattr(self, key)

    def get(self, key: str) -> Any | None:
        """Get a value by string and return None if not present."""
        return getattr(self, key) if hasattr(self, key) else None

    def get_key(self) -> RingEventKey:
        """Return the identificationkey for the event."""
        return RingEventKey(self.id, self.doorbot_id, self.kind, self.now)


class RingEventKey(NamedTuple):
    """Class to identify an event.

    Used for determining if messages are updates to events.
    """

    id: int
    doorbot_id: int
    kind: str
    now: float



"""Module for listening to firebase cloud messages and updating dings."""

from __future__ import annotations

import asyncio
import json
import logging
import time
from typing import TYPE_CHECKING, Any, Callable

from async_timeout import timeout as asyncio_timeout
from firebase_messaging import FcmPushClient, FcmRegisterConfig

from ring_doorbell.const import (
    API_URI,
    API_VERSION,
    DEFAULT_LISTEN_EVENT_EXPIRES_IN,
    FCM_API_KEY,
    FCM_APP_ID,
    FCM_PROJECT_ID,
    FCM_RING_SENDER_ID,
    KIND_DING,
    KIND_INTERCOM_UNLOCK,
    KIND_MOTION,
    PUSH_ACTION_DING,
    PUSH_ACTION_INTERCOM_UNLOCK,
    PUSH_ACTION_MOTION,
    PUSH_NOTIFICATION_KINDS,
    SUBSCRIPTION_ENDPOINT,
)
from ring_doorbell.event import RingEvent, RingEventKey
from ring_doorbell.exceptions import RingError
from ring_doorbell.util import parse_datetime

from .listenerconfig import RingEventListenerConfig

if TYPE_CHECKING:
    from ring_doorbell.ring import Ring

_logger = logging.getLogger(__name__)

OnNotificationCallable = Callable[[RingEvent], None]
CredentialsUpdatedCallable = Callable[[dict[str, Any]], None]


class RingEventListener:
    """Class to connect to firebase cloud messaging."""

    SESSION_REFRESH_INTERVAL = 60 * 60 * 12

    def __init__(
        self,
        ring: Ring,
        credentials: dict[str, Any] | None = None,
        credentials_updated_callback: CredentialsUpdatedCallable | None = None,
        *,
        config: RingEventListenerConfig | None = None,
    ) -> None:
        """Initialise the event listener with credentials.

        Provide a callback for when credentials are updated by FCM.
        """
        self._ring = ring

        self._callbacks: dict[int, OnNotificationCallable] = {}
        self.subscribed = False
        self.started = False
        self._device_model = self._ring.auth.get_device_model()

        self._credentials = credentials
        self._credentials_updated_callback = credentials_updated_callback

        self._receiver: FcmPushClient | None = None
        self._config: RingEventListenerConfig = (
            config or RingEventListenerConfig.default_config()
        )

        self._subscription_counter = 1
        self._intercom_unlock_counter: dict[int, int] = {}

        self.session_refresh_task: asyncio.Task | None = None
        self.fcm_token: str | None = None

        self._seen_events: set[RingEventKey] = set()

    def _credentials_updated_cb(self, creds: dict[str, Any]) -> None:
        self._credentials = creds
        if self._credentials_updated_callback:
            self._credentials_updated_callback(creds)

    async def add_subscription_to_ring(self, token: str) -> None:
        """Add subscription to ring."""
        if not self._ring.session:
            await self._ring.async_create_session()

        session_patch_data = {
            "device": {
                "metadata": {
                    "api_version": API_VERSION,
                    "device_model": self._device_model,
                    "pn_dict_version": "2.0.0",
                    "pn_service": "fcm",
                },
                "os": "android",
                "push_notification_token": token,
            }
        }
        resp = await self._ring.auth.async_query(
            API_URI + SUBSCRIPTION_ENDPOINT,
            method="PATCH",
            json=session_patch_data,
            raise_for_status=False,
        )
        if resp.status_code != 204:
            _logger.error(
                "Unable to checkin to listen service, "
                "response was %s %s, event listener not started",
                resp.status_code,
                resp.text,
            )
            self.subscribed = False
            return

        self.subscribed = True
        # Update devices for the intercom unlock events
        if not self._ring.devices_data:
            await self._ring.async_update_devices()

    def add_notification_callback(self, callback: OnNotificationCallable) -> int:
        """Add a callback to be notified on event."""
        sub_id = self._subscription_counter

        self._callbacks[sub_id] = callback
        self._subscription_counter += 1

        return sub_id

    def remove_notification_callback(self, subscription_id: int) -> None:
        """Remove a notification callback by id."""
        if subscription_id == 1:
            msg = "Cannot remove the default callback for ring-doorbell with value 1"
            raise RingError(msg)

        if subscription_id not in self._callbacks:
            msg = f"ID {subscription_id} is not a valid callback id"
            raise RingError(msg)

        del self._callbacks[subscription_id]

    async def stop(self) -> None:
        """Stop the listener."""
        self.started = False

        if self._receiver:
            await self._receiver.stop()

        refresh_task = self.session_refresh_task
        self.session_refresh_task = None
        if refresh_task and not refresh_task.done():
            refresh_task.cancel()

        self._callbacks = {}

    async def start(
        self,
        *,
        timeout: int = 10,
    ) -> bool:
        """Start the listener."""
        _logger.debug("Starting event listener")
        if not self._receiver:
            fcm_config = FcmRegisterConfig(
                FCM_PROJECT_ID, FCM_APP_ID, FCM_API_KEY, FCM_RING_SENDER_ID
            )
            self._receiver = FcmPushClient(
                self._on_notification,
                fcm_config,
                self._credentials,
                self._credentials_updated_cb,
                config=self._config,
                http_client_session=self._ring.auth._session,  # noqa: SLF001
            )
        self.fcm_token = await self._receiver.checkin_or_register()
        if not self.fcm_token:
            _logger.error(
                "Ring listener unable to check in to fcm, " "event listener not started"
            )
            return False

        if not self.subscribed:
            await self.add_subscription_to_ring(self.fcm_token)
        if self.subscribed:
            self.add_notification_callback(self._ring._add_event_to_dings_data)  # noqa: SLF001

            async with asyncio_timeout(timeout):
                await self._receiver.start()
            self.started = True
            self.session_refresh_task = asyncio.create_task(
                self._periodic_session_refresh()
            )
            _logger.debug("Started event listener")
        return self.started

    async def _periodic_session_refresh(self) -> None:
        while self.started:
            now = time.monotonic()
            if TYPE_CHECKING:
                assert self._ring.session_refresh_time
                assert self.fcm_token
            since_refresh = now - self._ring.session_refresh_time

            if since_refresh > self.SESSION_REFRESH_INTERVAL:
                _logger.debug("Refreshing ring session")
                await self._ring.async_create_session()
                await self.add_subscription_to_ring(self.fcm_token)
                break

            sleep_for = 1 + time.monotonic() - self._ring.session_refresh_time
            await asyncio.sleep(sleep_for)

    def _get_ding_event(self, gcm_data: dict[str, Any]) -> RingEvent:
        ding = gcm_data["ding"]
        action = gcm_data["action"]
        subtype = gcm_data["subtype"]
        if action.lower() == PUSH_ACTION_MOTION.lower():
            kind = KIND_MOTION
            state = subtype
        elif action.lower == PUSH_ACTION_DING.lower():
            kind = KIND_DING
            state = "ringing"
        else:
            kind = action
            state = subtype

        created_at = ding["created_at"]
        create_seconds = parse_datetime(created_at).timestamp()
        return RingEvent(
            id=ding["id"],
            kind=kind,
            doorbot_id=ding["doorbot_id"],
            device_name=ding["device_name"],
            device_kind=ding["device_kind"],
            now=create_seconds,
            expires_in=DEFAULT_LISTEN_EVENT_EXPIRES_IN,
            state=state,
        )

    def _get_intercom_unlock_event(self, gcm_data: dict[str, Any]) -> RingEvent | None:
        device_api_id = gcm_data["alarm_meta"]["device_zid"]
        if (device := self._ring.get_device_by_api_id(device_api_id)) is None:
            _logger.debug("Event received for unknown device id: %s", device_api_id)
            return None

        if device_api_id not in self._intercom_unlock_counter:
            self._intercom_unlock_counter[device_api_id] = 0
        self._intercom_unlock_counter[device_api_id] += 1
        return RingEvent(
            id=self._intercom_unlock_counter[device_api_id],
            kind=KIND_INTERCOM_UNLOCK,
            doorbot_id=device_api_id,
            device_name=device.name,
            device_kind=device.kind,
            now=time.time(),
            expires_in=DEFAULT_LISTEN_EVENT_EXPIRES_IN,
            state="unlock",
        )

    def _check_is_update(self, ring_event: RingEvent) -> None:
        """Battery doorbells send two events.

        First without an image and the second with an image.
        """
        now = time.time()
        seen_events = {
            key
            for key in self._seen_events
            if (now - key.now) < DEFAULT_LISTEN_EVENT_EXPIRES_IN
        }
        event_key = ring_event.get_key()
        if event_key in seen_events:
            ring_event.is_update = True
        else:
            seen_events.add(event_key)
        self._seen_events = seen_events

    def _on_notification(
        self,
        notification: dict[str, dict[str, str]],
        persistent_id: str,  # noqa: ARG002
        obj: Any | None = None,  # noqa: ARG002
    ) -> None:
        msg_data = notification["data"]
        if "gcmData" in msg_data:
            gcm_data = json.loads(notification["data"]["gcmData"])
            ring_event = self._get_legacy_ring_event(gcm_data)
        else:
            ring_event = self._get_ring_event(msg_data)

        if ring_event:
            self._check_is_update(ring_event)
            _logger.debug("Event received %s", ring_event)
            for callback in self._callbacks.values():
                callback(ring_event)
        else:
            _logger.debug("Unknown event received %s", msg_data)

    def _get_ring_event(self, msg_data: dict) -> RingEvent | None:
        if (android_config_str := msg_data.get("android_config")) is None or (
            data_str := msg_data.get("data")
        ) is None:
            _logger.debug(
                "Unexpected alert type in fcm message data.  Full message is:\n%s",
                json.dumps(msg_data),
            )
            return None

        android_config = json.loads(android_config_str)
        data = json.loads(data_str)
        event_category = android_config["category"]
        event_kind = PUSH_NOTIFICATION_KINDS.get(event_category, "Unknown")
        device = data["device"]
        event = data["event"]
        event_id = int(event["ding"]["id"])
        created_at = event["ding"]["created_at"]
        create_seconds = parse_datetime(created_at).timestamp()
        return RingEvent(
            event_id,
            device["id"],
            device_name=device.get("name"),
            device_kind=device.get("kind"),
            kind=event_kind,
            now=create_seconds,
            expires_in=DEFAULT_LISTEN_EVENT_EXPIRES_IN,
            state=event["ding"]["subtype"],
        )

    def _get_legacy_ring_event(self, gcm_data: dict) -> RingEvent | None:
        re: RingEvent | None = None
        if "ding" in gcm_data:
            re = self._get_ding_event(gcm_data)
        elif gcm_data.get("action") == PUSH_ACTION_INTERCOM_UNLOCK:
            re = self._get_intercom_unlock_event(gcm_data)
        elif "community_alert" not in gcm_data:
            _logger.debug(
                "Unexpected alert type in gcmData.  Full message is:\n%s",
                json.dumps(gcm_data),
            )
            return None
        return re



# vim:sw=4:ts=4:et:
"""Python Ring light group wrapper."""

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING, Any, ClassVar

from ring_doorbell.const import (
    GROUP_DEVICES_ENDPOINT,
    MSG_ALLOWED_VALUES,
    RingCapability,
)
from ring_doorbell.exceptions import RingError

_LOGGER = logging.getLogger(__name__)


class RingLightGroup:
    """Implementation for RingLightGroup."""

    if TYPE_CHECKING:
        from ring_doorbell.ring import Ring

    def __init__(self, ring: Ring, group_id: str) -> None:
        """Initialize Ring Light Group."""
        self._ring = ring
        self.group_id = group_id  # pylint:disable=invalid-name
        self._health_attrs: dict[str, Any] = {}
        self._health_attrs_fetched = False

    def __repr__(self) -> str:
        """Return __repr__."""
        return f"<{self.__class__.__name__}: {self.name}>"

    async def async_update(self) -> None:
        """Update this device info."""
        url = GROUP_DEVICES_ENDPOINT.format(self.location_id, self.group_id)
        resp = await self._ring.async_query(url)
        self._health_attrs = resp.json()
        self._health_attrs_fetched = True

    @property
    def _attrs(self) -> dict[str, Any]:
        """Return attributes."""
        return self._ring.groups_data[self.group_id]

    @property
    def id(self) -> str:
        """Return ID."""
        return self.group_id

    @property
    def name(self) -> str:
        """Return name."""
        return self._attrs["name"]

    @property
    def family(self) -> str:
        """Return Ring device family type."""
        return "group"

    @property
    def device_id(self) -> str:
        """Return group ID. Deprecated."""
        warnings.warn(
            "RingLightGroup.device_id is deprecated; use group_id",
            DeprecationWarning,
            stacklevel=1,
        )
        return self.group_id

    @property
    def location_id(self) -> str:
        """Return group location ID."""
        return self._attrs["location_id"]

    @property
    def model(self) -> str:
        """Return Ring device model name."""
        return "Light Group"

    def has_capability(self, capability: RingCapability | str) -> bool:
        """Return if device has specific capability."""
        capability = (
            capability
            if isinstance(capability, RingCapability)
            else RingCapability.from_name(capability)
        )
        return capability == RingCapability.LIGHT

    @property
    def lights(self) -> bool:
        """Return lights status."""
        if not self._health_attrs_fetched:
            msg = (
                "You need to call update on the "
                "group before accessing the lights property."
            )
            raise RingError(msg)
        return self._health_attrs["lights_on"]

    async def async_set_lights(
        self,
        state: bool | tuple[bool, int],
        duration: int | None = None,
    ) -> None:
        """Control the lights."""
        values = ["True", "False"]
        if isinstance(state, tuple):
            state, duration = state

        if not isinstance(state, bool):
            raise RingError(MSG_ALLOWED_VALUES.format(", ".join(values)))

        url = GROUP_DEVICES_ENDPOINT.format(self.location_id, self.group_id)
        payload: dict[str, dict[str, bool | int]] = {"lights_on": {"enabled": state}}
        if duration is not None:
            payload["lights_on"]["duration_seconds"] = duration
        await self._ring.async_query(url, method="POST", json=payload)
        await self.async_update()

    DEPRECATED_API_QUERIES: ClassVar = {
        "update",
    }
    DEPRECATED_API_PROPERTY_SETTERS: ClassVar = {
        "lights",
    }

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            """Get a deprecated attribute or raise an error."""
            if name in self.DEPRECATED_API_QUERIES:
                return self._ring.auth._dep_handler.get_api_query(self, name)  # noqa: SLF001
            msg = f"{self.__class__.__name__} has no attribute {name!r}"
            raise AttributeError(msg)

        def __setattr__(self, name: str, value: Any) -> None:
            """Set a deprecated attribute or raise an error."""
            if name in self.DEPRECATED_API_PROPERTY_SETTERS:
                self._ring.auth._dep_handler.set_api_property(self, name, value)  # noqa: SLF001
            else:
                super().__setattr__(name, value)



# vim:sw=4:ts=4:et:
"""Python Ring Other (Intercom) wrapper."""

from __future__ import annotations

import json
import logging
import uuid
from typing import TYPE_CHECKING, Any, ClassVar

from ring_doorbell.const import (
    DOORBELLS_ENDPOINT,
    HEALTH_DOORBELL_ENDPOINT,
    INTERCOM_ALLOWED_USERS,
    INTERCOM_INVITATIONS_DELETE_ENDPOINT,
    INTERCOM_INVITATIONS_ENDPOINT,
    INTERCOM_KINDS,
    INTERCOM_OPEN_ENDPOINT,
    MIC_VOL_MAX,
    MIC_VOL_MIN,
    MSG_VOL_OUTBOUND,
    OTHER_DOORBELL_VOL_MAX,
    OTHER_DOORBELL_VOL_MIN,
    SETTINGS_ENDPOINT,
    VOICE_VOL_MAX,
    VOICE_VOL_MIN,
    RingCapability,
)
from ring_doorbell.exceptions import RingError
from ring_doorbell.generic import RingGeneric

_LOGGER = logging.getLogger(__name__)


class RingOther(RingGeneric):
    """Implementation for Ring Intercom."""

    if TYPE_CHECKING:
        from ring_doorbell.ring import Ring

    def __init__(self, ring: Ring, device_api_id: int, *, shared: bool = False) -> None:
        """Initialise the other devices."""
        super().__init__(ring, device_api_id)
        self.shared = shared

    @property
    def family(self) -> str:
        """Return Ring device family type."""
        return "other"

    async def async_update_health_data(self) -> None:
        """Update health attrs."""
        resp = await self._ring.async_query(
            HEALTH_DOORBELL_ENDPOINT.format(self.device_api_id)
        )
        self._health_attrs = resp.json().get("device_health", {})

    @property
    def model(self) -> str:
        """Return Ring device model name."""
        if self.kind in INTERCOM_KINDS:
            return "Intercom"
        return "Unknown Other"

    def has_capability(self, capability: RingCapability | str) -> bool:
        """Return if device has specific capability."""
        capability = (
            capability
            if isinstance(capability, RingCapability)
            else RingCapability.from_name(capability)
        )
        if capability in [
            RingCapability.OPEN,
            RingCapability.HISTORY,
            RingCapability.DING,
        ]:
            return self.kind in INTERCOM_KINDS
        return False

    @property
    def battery_life(self) -> int | None:
        """Return battery life."""
        if self.kind in INTERCOM_KINDS:
            if self._attrs.get("battery_life") is None:
                return None

            value = int(self._attrs.get("battery_life", 0))
            if value and value > 100:
                value = 100

            return value
        return None

    @property
    def subscribed(self) -> bool:
        """Return if is online."""
        if self.kind in INTERCOM_KINDS:
            result = self._attrs.get("subscribed")
            return result is not None
        return False

    @property
    def has_subscription(self) -> bool:
        """Return boolean if the account has subscription."""
        if self.kind in INTERCOM_KINDS and (features := self._attrs.get("features")):
            return features.get("show_recordings", False)
        return False

    @property
    def unlock_duration(self) -> str | None:
        """Return time unlock switch is held closed."""
        return (
            json.loads(self._attrs["settings"]["intercom_settings"]["config"])
            .get("analog", {})
            .get("unlock_duration")
        )

    @property
    def doorbell_volume(self) -> int:
        """Return doorbell volume."""
        if self.kind in INTERCOM_KINDS:
            return self._attrs["settings"].get("doorbell_volume", 0)
        return 0

    async def async_set_doorbell_volume(self, value: int) -> None:
        """Set the doorbell volume."""
        if not (
            (isinstance(value, int))
            and (OTHER_DOORBELL_VOL_MIN <= value <= OTHER_DOORBELL_VOL_MAX)
        ):
            raise RingError(
                MSG_VOL_OUTBOUND.format(OTHER_DOORBELL_VOL_MIN, OTHER_DOORBELL_VOL_MAX)
            )

        params = {
            "doorbot[settings][doorbell_volume]": str(value),
        }
        url = DOORBELLS_ENDPOINT.format(self.device_api_id)
        await self._ring.async_query(url, extra_params=params, method="PUT")

    @property
    def keep_alive_auto(self) -> float | None:
        """The keep alive auto setting."""
        if self.kind in INTERCOM_KINDS:
            return self._attrs["settings"].get("keep_alive_auto")
        return None

    async def async_set_keep_alive_auto(self, value: float) -> None:
        """Update the keep alive auto setting."""
        url = SETTINGS_ENDPOINT.format(self.device_api_id)
        payload = {"keep_alive_settings": {"keep_alive_auto": value}}

        await self._ring.async_query(url, method="PATCH", json=payload)

    @property
    def mic_volume(self) -> int | None:
        """Return mic volume."""
        if self.kind in INTERCOM_KINDS:
            return self._attrs["settings"].get("mic_volume")
        return None

    async def async_set_mic_volume(self, value: int) -> None:
        """Set the mic volume."""
        if not ((isinstance(value, int)) and (MIC_VOL_MIN <= value <= MIC_VOL_MAX)):
            raise RingError(MSG_VOL_OUTBOUND.format(MIC_VOL_MIN, MIC_VOL_MAX))

        url = SETTINGS_ENDPOINT.format(self.device_api_id)
        payload = {"volume_settings": {"mic_volume": value}}

        await self._ring.async_query(url, method="PATCH", json=payload)

    @property
    def voice_volume(self) -> int | None:
        """Return voice volume."""
        if self.kind in INTERCOM_KINDS:
            return self._attrs["settings"].get("voice_volume")
        return None

    async def async_set_voice_volume(self, value: int) -> None:
        """Set the voice volume."""
        if not ((isinstance(value, int)) and (VOICE_VOL_MIN <= value <= VOICE_VOL_MAX)):
            raise RingError(MSG_VOL_OUTBOUND.format(VOICE_VOL_MIN, VOICE_VOL_MAX))

        url = SETTINGS_ENDPOINT.format(self.device_api_id)
        payload = {"volume_settings": {"voice_volume": value}}

        await self._ring.async_query(url, method="PATCH", json=payload)

    async def async_get_clip_length_max(self) -> int | None:
        """Get the Maximum clip length."""
        url = SETTINGS_ENDPOINT.format(self.device_api_id)
        resp = await self._ring.async_query(url, method="GET")
        return resp.json().get("video_settings", {}).get("clip_length_max")

    async def async_set_clip_length_max(self, value: int) -> None:
        """Set the maximum clip length.

        This value sets an effective refractory period on consecutive rigns
        eg if set to default value of 60, rings occuring with 60 seconds of
        first will not be detected.
        """
        url = SETTINGS_ENDPOINT.format(self.device_api_id)
        payload = {"video_settings": {"clip_length_max": value}}
        await self._ring.async_query(url, method="PATCH", json=payload)

    @property
    def connection_status(self) -> str | None:
        """Return connection status."""
        if self.kind in INTERCOM_KINDS:
            return self._attrs.get("alerts", {}).get("connection")
        return None

    async def async_get_allowed_users(self) -> list[dict[str, Any]] | None:
        """Return list of users allowed or invited to access."""
        if self.kind in INTERCOM_KINDS:
            url = INTERCOM_ALLOWED_USERS.format(self.location_id)
            resp = await self._ring.async_query(url, method="GET")
            return resp.json()

        return None

    async def async_open_door(self, user_id: int = -1) -> bool:
        """Open the door."""
        if self.kind in INTERCOM_KINDS:
            url = INTERCOM_OPEN_ENDPOINT.format(self.device_api_id)
            request_id = str(uuid.uuid4())
            # params can also accept:
            # issue_time: in seconds
            # command_timeout: in seconds
            payload = {
                "command_name": "device_rpc",
                "request": {
                    "id": request_id,
                    "jsonrpc": "2.0",
                    "method": "unlock_door",
                    "params": {
                        "door_id": 0,
                        "user_id": user_id,
                    },
                },
            }
            resp = await self._ring.async_query(url, method="PUT", json=payload)
            response = resp.json()
            if response.get("result", {}).get("code", -1) == 0:
                return True

        return False

    async def async_invite_access(self, email: str) -> bool:
        """Invite user."""
        if self.kind in INTERCOM_KINDS:
            url = INTERCOM_INVITATIONS_ENDPOINT.format(self.location_id)
            payload = {
                "invitation": {
                    "doorbot_ids": [self.device_api_id],
                    "invited_email": email,
                    "group_ids": [],
                }
            }
            await self._ring.async_query(url, method="POST", json=payload)
            return True

        return False

    async def async_remove_access(self, user_id: int) -> bool:
        """Remove user access or invitation."""
        if self.kind in INTERCOM_KINDS:
            url = INTERCOM_INVITATIONS_DELETE_ENDPOINT.format(self.location_id, user_id)
            await self._ring.async_query(url, method="DELETE")
            return True

        return False

    DEPRECATED_API_QUERIES: ClassVar = {
        *RingGeneric.DEPRECATED_API_QUERIES,
        "update_health_data",
        "open_door",
        "invite_access",
        "remove_access",
    }
    DEPRECATED_API_PROPERTY_GETTERS: ClassVar = {
        *RingGeneric.DEPRECATED_API_PROPERTY_GETTERS,
        "clip_length_max",
        "allowed_users",
    }
    DEPRECATED_API_PROPERTY_SETTERS: ClassVar = {
        *RingGeneric.DEPRECATED_API_PROPERTY_SETTERS,
        "doorbell_volume",
        "keep_alive_auto",
        "mic_volume",
        "voice_volume",
        "clip_length_max",
    }
